/**
 *  joola.io
 *
 *  Copyright Joola Smart Solutions, Ltd. <info@joo.la>
 *
 *  Licensed under GNU General Public License 3.0 or later.
 *  Some rights reserved. See LICENSE, AUTHORS.
 *
 *  @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>
 */

var
  router = require('./index'),
  url = require('url'),

  _datasources = require('../lib/objects/datasources'),
  connector = require('../lib/connectors/connector');

/**
 * Lists all available dimensions. You can provide parameters and ask for a specific data source and datatable.
 *
 * The returned list will contain only dimensions authorized for the authenticated user.
 *
 * @method list
 * @param {string} [datasourceid] Datasource ID.
 * @param {string} [datatableid] Datatable ID.
 * @return {object} json structure listing all dimensions.
 */
exports.list = {
  name: 'datasources/list',
  description: 'i list all datasources available',
  inputs: {
    required: [],
    optional: ['datasourceid']
  },
  outputExample: {},
  permission: ['access_system'],
  run: function (req, res) {
    var response = {};
    var datasources = _datasources.list();
    if (!datasources)
      return router.responseError(new router.ErrorTemplate('Failed to find data table'), req, res);

    response.datatables = ce.clone(datasources);
    return router.responseSuccess(response, req, res);

  }
};

/**
 * Lists all available dimensions. You can provide parameters and ask for a specific data source and datatable.
 *
 * The returned list will contain only dimensions authorized for the authenticated user.
 *
 * @method get
 * @param {string} [dimensionid] Dimension ID.
 * @return {object} json structure holding the dimension details.
 */
exports.get = {
  name: 'datasources/get',
  description: 'i get a specific datasources',
  inputs: {
    required: ['id'],
    optional: []
  },
  outputExample: {},
  permission: ['access_system'],
  run: function (req, res) {
    var response = {};
    var datasource = _datasources.get(req.params.datasourceid);
    if (!datasource)
      return router.responseError(new router.ErrorTemplate('Failed to find data table'), req, res);

    response.datasource = ce.cloneuptolevel(datasource, 0);
    return router.responseSuccess(response, req, res);
  }
};

/**
 * Validates the dimension database integrity.
 *
 * @method get
 * @param {string} [dimensionid] Dimension ID.
 * @return {object} json structure holding the validation details.
 */
exports.validate = {
  name: 'datasources/validate',
  description: 'i validate a specific datasources',
  inputs: {
    required: ['datasourceid'],
    optional: []
  },
  outputExample: {},
  permission: ['manage_system'],
  run: function (req, res) {
    var response = {};
    var datasource = _datasources.get(req.params.datasourceid);
    var query = connector.createQuery();

    response.datasource = ce.clone(datasource);
    response.validated = false;

    query.sql = 'SELECT 1;';
    query.limit = 1;
    query.datasource = datasource;
    connector.executeQuery(query, function (query, rows, fields, error) {
      response.queryInformation = ce.clone(query);
      if (error)
        throw error;
      else {
        response.validated = true;
      }
      return router.responseSuccess(response, req, res);
    });
  }
};
